//
//	servo motor handling test program for PIC16F1827
//	Copyright JL1VNQ / HARU
//
//	servo motor: PWM period 15-20ms, pulse width  1-2ms (0.5-2.5ms)
//
//
//

#define EEPROM_SIZE 	256
#define _XTAL_FREQ		4000000

#include <xc.h>
#include <stdio.h>

//for 16F1827
#pragma config CPD = OFF, BOREN = NSLEEP, IESO = OFF, FOSC = INTOSC
#pragma	config FCMEN = OFF, MCLRE = OFF, WDTE = OFF, CP = OFF, PWRTE = ON, CLKOUTEN = OFF
#pragma config PLLEN = OFF, WRT = OFF, STVREN = OFF, BORV = LO, LVP = OFF

#define SDA 	    LATAbits.LATA2
#define SCL			LATAbits.LATA3

#define CONT		LATAbits.LATA4					// motor control pwm output

#define RE_A		PORTB,2
#define RE_B		PORTB,3

#define LCD_AD		0x7C							// AQM0802 I2C address

#define TMR0_set	0x83							// TMR0 1msec interval


//__EEPROM_DATA(0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00);		// (reserved)

unsigned char change_flag = 0;						// position change flag
unsigned int offset;								// pulse width offset
unsigned char lcdbuf[9];							// lcd print buffer
unsigned int pulse_width;
unsigned char buffH;
unsigned char buffL;

void msec_delay(unsigned short time);
void I2C_init(void);
void I2C_send(unsigned char data);
void lcd_cmd(unsigned char work);
void lcd_data(unsigned char work);
void lcd_init(void);
void lcd_position(unsigned char li, unsigned char col);
void lcd_str_disp(unsigned char li, unsigned char col, const char *string);
void lcd_clear(void);
void offset_disp_set(void);
void cgram_set(void);
void bar_disp(void);

static void interrupt sw_sta(void){
	if(PIR1bits.TMR2IF){
		INTCONbits.GIE = 0;
		TMR2IF = 0;

		CONT = 1;									// ON

		INTCONbits.TMR0IE = 0;

		TMR1H = buffH;								// TMR1(OFF timer) value set(variable)
		TMR1L = buffL;

		PR2 = 124;									// TMR2(ON timer) value set(160us x 125 = 20ms fixed)
		TMR2 = 0;

		INTCONbits.GIE = 1;
	}

	else if(PIR1bits.TMR1IF){
		INTCONbits.GIE = 0;
		TMR1IF = 0;

		CONT = 0;									// OFF

		INTCONbits.TMR0IE = 1;

		INTCONbits.GIE = 1;
	}

	if(INTCONbits.TMR0IF){
		INTCONbits.GIE = 0;
		TMR0IF = 0;

		static unsigned char re_sta = 0xFC;			// rotary encoder status
		static unsigned char up_keep = 0, down_keep = 0;		// ON state flag

		re_sta = ((PORTB & 0x0C) | (re_sta << 2)) & 0xFC;		// rotary encoder status 3points 1masked

		if(re_sta == 0xE8) up_keep = 1;				// RE_A(UP) fall edge detect(111010xx)
		else if(re_sta == 0xD4) down_keep = 1;		// RE_B(DOWN) fall edge detect(110101xx)

		if(up_keep && re_sta == 0x14){				// RE_A(UP) rise edge detect(000101xx)
			up_keep = 0;
			if(offset > 439) offset = 440;
			else offset++;
			change_flag = 1;
		}
		else if(down_keep && re_sta == 0x28){		// RE_B(DOWN) rise edge detect(001010xx)
			down_keep = 0;
			if(offset < 1) offset = 0;
			else offset--;
			change_flag = 1;
		}

		TMR0 = TMR0_set;
		INTCONbits.GIE = 1;
	}
}

void main(void){

	OSCCON = 0x6A;									// 4MHz internal OSC no PLL

	ANSELA = 0x00;
	ANSELB = 0x00;
	TRISA = 0x00;
	TRISB = 0x0C;
	WPUB = 0x0C;									// PORTB weak pull-up

	PORTA &= 0x70;									// PORTA initialize

	OPTION_REG = 0x02;								// weak pull-up, TMR0 internal clock(1us/cycle), prescaler 1:8

	I2C_init();
	lcd_init();

	cgram_set();

	lcd_str_disp(0,0,"ServoTst");					// start splash
	lcd_str_disp(1,0,"20150409");
	msec_delay(1000);
	lcd_clear();
//	lcd_str_disp(0,0,"DutyCycl");
	lcd_position(0,0);
	lcd_data(0x00);
	lcd_data(0x01);
	lcd_data(0x02);
	lcd_data(0x03);

	TMR0 = TMR0_set;								// 1msec
	INTCONbits.TMR0IF = 0;
	INTCONbits.TMR0IE = 1;

	T1CON = 0x21;									// TMR1 setting: FOSC/4(1us), prescaler 1:4(4us/count), TMR1 enable
	TMR1H = 0xFE;									// 1200us(300counts), 10000h - 12Ch
	TMR1L = 0xD4;
	PIR1bits.TMR1IF = 0;
	PIE1bits.TMR1IE = 1;

	T2CON = 0x4E;									// TMR2 setting: (FOSC/4) postscaler 1:10, TNR2 enable, prescaler 1:16
	PR2 = 124;										// 160us x 125 = 20ms
	TMR2 = 0;
	PIR1bits.TMR2IF = 0;
	PIE1bits.TMR2IE = 1;

	INTCONbits.PEIE = 1;

	INTCONbits.GIE = 1;

	offset = 220;									// neutral position
	bar_disp();
	offset_disp_set();

	while(1){
		if(change_flag == 1){
			INTCONbits.GIE = 0;
			offset_disp_set();
			change_flag = 0;
			INTCONbits.GIE = 1;
		}
	}
}

void msec_delay(unsigned short time){
	unsigned short i;
	for(i=0;i<time;i++){
		__delay_ms(1);
	}
}

void I2C_init(void){
	SDA = 1;
	SCL = 1;
}

void lcd_init(void){
	lcd_cmd(0x38);
	lcd_cmd(0x39);
	lcd_cmd(0x14);
	lcd_cmd(0x73);									// contrast value
//	lcd_cmd(0x56);									// VDD 3.3V
	lcd_cmd(0x52);									// VDD 5V
	lcd_cmd(0x6C);
	msec_delay(210);
	lcd_cmd(0x38);
	lcd_cmd(0x0C);
	lcd_cmd(0x01);
	msec_delay(2);
}

void I2C_send(unsigned char data){
	unsigned char i;
	for(i=0;i<8;i++){
		if(data & 0x80)
			SDA = 1;
//		else
//			SDA = 0;
		SCL = 1;
		SCL = 0;
		SDA = 0;
		data <<= 1;
	}
	SCL = 1;										// for ack
	SCL = 0;
}

void lcd_cmd(unsigned char work){
	SDA = 0;
	SCL = 0;
	I2C_send(LCD_AD);
	I2C_send(0x80);									// Co=1, RS=0
	I2C_send(work);
	SCL = 1;
	SDA = 1;
	__delay_us(30);
}

void lcd_data(unsigned char work){
	SDA = 0;
	SCL = 0;
	I2C_send(LCD_AD);
	I2C_send(0xC0);									// Co=1, RS=1
	I2C_send(work);
	SCL = 1;
	SDA = 1;
	__delay_us(30);
}

void lcd_position(unsigned char li, unsigned char col){
	lcd_cmd(0x80 | (li << 6) | col);
}

void lcd_str_disp(unsigned char li, unsigned char col, const char *string){
	unsigned char i = 0;
	lcd_position(li,col);
	while(((col + i) < 8) && string[i]){
		lcd_data(string[i]);
		i++;
	}
}

void lcd_clear(void){
	lcd_cmd(0x01);
	msec_delay(2);
}

void offset_disp_set(void){
	bar_disp();
	sprintf(lcdbuf,"Deg=%3d",(offset * 18 / 44));	// offset value to degree convert
	lcd_str_disp(1,0,lcdbuf);
	lcd_data(0xDF);
	pulse_width = 64944 + offset;
	buffH = pulse_width >> 8;
	buffL = pulse_width & 0x00FF;
}

void cgram_set(void){
	unsigned char i;
	for(i=0;i<7;i++){
		lcd_cmd(0x40 + i);
		if(i == 3) lcd_data(0x04);
		else lcd_data(0x00);
		lcd_cmd(0x48 + i);
		if(i == 3) lcd_data(0x14);
		else lcd_data(0x10);
		lcd_cmd(0x50 + i);
		lcd_data(0x14);
		lcd_cmd(0x58 + i);
		lcd_data(0x15);
	}
}

void bar_disp(void){
	unsigned char i;
	lcd_position(0,0);
	lcd_data('L');
	if(offset < 75){
		if(offset == 0) lcd_data(0);
		else if(offset < 25) lcd_data(1);
		else if(offset < 50) lcd_data(2);
		else lcd_data(3);	
		for(i=0;i<5;i++){
			lcd_data(0);
		}
	}
	else if(offset < 150){
		lcd_data(3);
		if(offset < 100) lcd_data(1);
		else if(offset <125) lcd_data(2);
		else lcd_data(3);
		for(i=0;i<4;i++){
			lcd_data(0);
		}
	}
	else if(offset < 225){
		lcd_data(3);
		lcd_data(3);
		if(offset < 175) lcd_data(1);
		else if(offset <200) lcd_data(2);
		else lcd_data(3);
		for(i=0;i<3;i++){
			lcd_data(0);
		}
	}
	else if(offset < 300){
		for(i=0;i<3;i++){
			lcd_data(3);
		}
		if(offset < 250) lcd_data(1);
		else if(offset <275) lcd_data(2);
		else lcd_data(3);
		lcd_data(0);
		lcd_data(0);
	}
	else if(offset < 375){
		for(i=0;i<4;i++){
			lcd_data(3);
		}
		if(offset < 325) lcd_data(1);
		else if(offset <350) lcd_data(2);
		else lcd_data(3);
		lcd_data(0);
	}
	else if(offset < 450){
		for(i=0;i<5;i++){
			lcd_data(3);
		}
		if(offset < 400) lcd_data(1);
		else if(offset <425) lcd_data(2);
		else lcd_data(3);
	}
	lcd_data('H');
}